"""
For each directory containing YAML rules, run those rules on the file in the same directory with the same name but different extension.
E.g. eqeq.yaml runs on eqeq.py.
Validate that the output is annotated in the source file with by looking for a comment like:

 ```
 # ruleid:eqeq-is-bad
 ```
 On the preceeding line.

 """
import argparse
import collections
import json
import sys
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple

from semgrep.constants import YML_EXTENSIONS
from semgrep.semgrep_main import main as semgrepmain
from semgrep.util import debug_print
from semgrep.util import print_error_exit


def normalize_rule_id(line: str) -> str:
    """
    given a line like `     # ruleid:foobar`
    or `      // ruleid:foobar`
    return `foobar`
    """
    return line.strip().split(":")[1].strip()


def compute_confusion_matrix(
    reported: Set[Any], expected: Set[Any]
) -> Tuple[int, int, int, int]:
    true_positives = len(expected.intersection(reported))
    false_positives = len(reported - expected)
    true_negatives = 0  # we have no way to label "ok"
    false_negatives = len(expected - reported)

    return (true_positives, true_negatives, false_positives, false_negatives)


def _test_compute_confusion_matrix() -> None:
    tp, tn, fp, fn = compute_confusion_matrix(set([1, 2, 3, 4]), set([1]))
    assert tp == 1
    assert tn == 0
    assert fp == 3
    assert fn == 0

    tp, tn, fp, fn = compute_confusion_matrix(set([1, 2, 3, 4]), set([1, 2, 3, 4]))
    assert tp == 4
    assert tn == 0
    assert fp == 0
    assert fn == 0

    tp, tn, fp, fn = compute_confusion_matrix(set([2, 3]), set([1, 2, 3, 4]))
    assert tp == 2
    assert tn == 0
    assert fp == 0
    assert fn == 2


def line_has_todo_rule(line: str) -> bool:
    return (
        "#todoruleid:" in line
        or "# todoruleid:" in line
        or "// todoruleid:" in line
        or "//todoruleid:" in line
    )


def line_has_rule(line: str) -> bool:
    return (
        "#ruleid:" in line
        or "# ruleid:" in line
        or "//ruleid:" in line
        or "// ruleid:" in line
    )


def line_has_todo_ok(line: str) -> bool:
    return (
        "#todook" in line
        or "# todook" in line
        or "// todook" in line
        or "//todook" in line
    )


def score_output_json(
    json_out: Dict[str, Any], test_files: List[Path], ignore_todo: bool
) -> Tuple[Dict[str, List[int]], Dict[str, Dict[str, Any]], int]:
    comment_lines: Dict[str, Dict[str, List[int]]] = collections.defaultdict(
        lambda: collections.defaultdict(list)
    )
    reported_lines: Dict[str, Dict[str, List[int]]] = collections.defaultdict(
        lambda: collections.defaultdict(list)
    )
    ignore_lines: Dict[str, List[int]] = collections.defaultdict(list)
    score_by_checkid: Dict[str, List[int]] = collections.defaultdict(
        lambda: [0, 0, 0, 0]
    )
    expected_reported_by_check_id: Dict[str, Dict[str, Any]] = collections.defaultdict(
        dict
    )
    num_todo = 0

    for test_file in test_files:
        test_file_resolved = str(test_file.resolve())
        with open(test_file_resolved) as fin:
            all_lines = fin.readlines()
            for i, line in enumerate(all_lines):
                # +1 because we are 0 based and semgrep output is not, plus skip the comment line
                effective_line_num = i + 2

                todo_in_line = line_has_todo_rule(line)
                todo_ok_in_line = line_has_todo_ok(line)
                if todo_in_line:
                    num_todo += 1
                if (not ignore_todo and todo_in_line) or line_has_rule(line):
                    comment_lines[test_file_resolved][normalize_rule_id(line)].append(
                        effective_line_num
                    )
                if ignore_todo and todo_ok_in_line:
                    ignore_lines[test_file_resolved].append(effective_line_num)

    for result in json_out["results"]:
        reported_lines[str(Path(result["path"]).resolve())][result["check_id"]].append(
            int(result["start"]["line"])
        )

    def join_keys(a: Dict[str, Any], b: Dict[str, Any]) -> Set[str]:
        return set(a.keys()).union(set(b.keys()))

    for file_path in join_keys(comment_lines, reported_lines):
        for check_id in join_keys(comment_lines[file_path], reported_lines[file_path]):
            all_reported = set(reported_lines[file_path][check_id])
            expected = set(comment_lines[file_path][check_id])
            ignored = set(ignore_lines[file_path])

            reported = all_reported - ignored

            new_cm = compute_confusion_matrix(reported, expected)
            debug_print(
                f"reported lines for check {check_id}: {sorted(reported)}, expected lines: {sorted(expected)} (ignored: {sorted(ignored)}, confusion matrix: {new_cm}"
            )
            expected_reported_by_check_id[check_id][file_path] = (expected, reported)
            # TODO: -- re-enable this
            # assert len(set(reported_lines[file_path][check_id])) == len(
            #    reported_lines[file_path][check_id]
            # ), f"for testing, please don't make rules that fire multiple times on the same line ({check_id} in {file_path} on lines {reported_lines[file_path][check_id]})"
            old_cm = score_by_checkid[check_id]
            score_by_checkid[check_id] = [
                old_cm[i] + new_cm[i] for i in range(len(new_cm))
            ]

    return (score_by_checkid, expected_reported_by_check_id, num_todo)


def confusion_matrix_to_string(confusion: List[int]) -> str:
    tp, tn, fp, fn = confusion[0], confusion[1], confusion[2], confusion[3]
    return f"TP: {tp}\tTN:{tn}\t FP: {fp}\t FN: {fn}"


def invoke_semgrep(
    strict: bool, test_files: List[Path], config: Path, unsafe: bool
) -> Any:
    return json.loads(
        semgrepmain(
            target=[str(t) for t in test_files],
            pattern="",
            lang="",
            config=str(config),
            generate_config=False,
            no_rewrite_rule_ids=True,
            jobs=1,
            include=[],
            include_dir=[],
            exclude=[],
            exclude_dir=[],
            exclude_tests=False,
            json_format=True,
            sarif=False,
            output_destination="",
            quiet=True,
            strict=strict,
            exit_on_error=False,
            autofix=False,
            dangerously_allow_arbitrary_code_execution_from_rules=unsafe,
        )
    )


def generate_file_pairs(
    location: Path, ignore_todo: bool, strict: bool, semgrep_verbose: bool, unsafe: bool
) -> None:
    filenames = list(location.rglob("*"))
    no_tests = []
    tested = []
    semgrep_error = []
    print("starting tests...")
    for filename in filenames:
        if (
            filename.suffix in YML_EXTENSIONS
            and not filename.name.startswith(".")
            and not filename.parent.name.startswith(".")
        ):
            # find all filenames that have the same name but not extension, or are in a folder with the same name as a the yaml file
            yaml_file_name_without_ext = filename.with_suffix("")

            children_test_files = [
                p
                for p in filenames
                if str(p.with_suffix("")) == (str(yaml_file_name_without_ext))
            ]
            # remove yaml files from the test lists
            test_files = [
                path
                for path in children_test_files
                if path.suffix not in YML_EXTENSIONS and path.is_file()
            ]
            if not len(test_files):
                no_tests.append(filename)
                continue
            # invoke semgrep
            try:
                output_json = invoke_semgrep(strict, test_files, filename, unsafe)
                tested.append(
                    (filename, score_output_json(output_json, test_files, ignore_todo))
                )
            except Exception as ex:
                print(
                    f"semgrep error running with config {filename} on {test_files}:\n{ex}"
                )
                semgrep_error.append(filename)

    if len(semgrep_error) and strict:
        print("exiting due to semgrep/config errors and strict flag")
        sys.exit(1)

    print(f"{len(no_tests)} yaml files missing tests")
    debug_print(f"missing tests: {no_tests}")
    print(f"{len(tested)} yaml files tested")
    print("check id scoring:")
    print("=" * 80)
    failed_tests = []
    total_confusion = [0, 0, 0, 0]

    for (filename, (output, expected_reported_by_check_id, num_todo)) in tested:
        print(filename)
        if not len(output.items()):
            print(f"  no checks fired (TODOs: {num_todo})")
        for check_id, (tp, tn, fp, fn) in output.items():
            good = (fp == 0) and (fn == 0)
            if not good:
                failed_tests.append(
                    (filename, check_id, expected_reported_by_check_id[check_id])
                )
            status = "✔" if good else "✖"
            todo_text = f"(TODOs: {num_todo})" if num_todo > 0 else ""
            confusion = [tp, tn, fp, fn]
            # add to the total confusion matrix
            total_confusion = [
                total_confusion[i] + confusion[i] for i in range(len(confusion))
            ]
            print(
                f"  {status} - {check_id.ljust(60)}{confusion_matrix_to_string(confusion)} {todo_text}"
            )

    print("=" * 80)
    print(f"final confusion matrix: {confusion_matrix_to_string(total_confusion)}")
    print("=" * 80)

    if len(failed_tests) > 0:
        print(f"failing rule files: ")
        for (filename, check_id, failed_test_files) in failed_tests:
            print(f" ✖ FAILED rule file: {filename} check: {check_id}")
            for test_file_path, (expected, reported) in failed_test_files.items():
                print(
                    f"              in test: {test_file_path}, expected lines: {sorted(expected)} != reported: {sorted(reported)}"
                )
        print(
            f"{len(failed_tests)} checks failed tests (run with verbose flag for more details)"
        )
        sys.exit(1)
    else:
        print("all tests passed")
        sys.exit(0)


def main(
    location: Path,
    ignore_todo: bool,
    verbose: bool,
    strict: bool,
    semgrep_verbose: bool,
    unsafe: bool,
) -> None:
    generate_file_pairs(location, ignore_todo, strict, semgrep_verbose, unsafe)


def test_main(args: argparse.Namespace) -> None:
    _test_compute_confusion_matrix()
    if len(args.target) != 1:
        print_error_exit("only one target directory allowed for tests")
    target = Path(args.target[0])
    main(
        target,
        args.test_ignore_todo,
        args.verbose,
        args.strict,
        args.verbose,
        args.dangerously_allow_arbitrary_code_execution_from_rules,
    )



from typing import Optional


class SemgrepException(Exception):
    """
    Parent class of all exceptions we anticipate in Semgrep commands

    All Semgrep Exceptions are caught and their error messages
    are displayed to the user.
    """

    def __init__(self) -> None:
        self.msg: Optional[str] = None
        self.code = 1


class OutdatedPythonException(SemgrepException):
    def __init__(self) -> None:
        super().__init__()
        self.msg = "Semgrep requires Python 3.5+. Please ensure you are using Python3.5+ to run semgrep."



#!/usr/bin/env python3
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from unittest.mock import MagicMock
from unittest.mock import PropertyMock

from semgrep.constants import RCE_RULE_FLAG
from semgrep.evaluation import enumerate_patterns_in_boolean_expression
from semgrep.evaluation import evaluate_expression as raw_evaluate_expression
from semgrep.pattern_match import PatternMatch
from semgrep.rule import Rule
from semgrep.semgrep_types import BooleanRuleExpression
from semgrep.semgrep_types import Operator
from semgrep.semgrep_types import OPERATORS
from semgrep.semgrep_types import PatternId
from semgrep.semgrep_types import Range


def evaluate_expression(
    exprs: List[BooleanRuleExpression],
    pattern_ids_to_pattern_matches: Dict[PatternId, List[PatternMatch]],
    flags: Optional[Dict[str, Any]] = None,
) -> Set[Range]:
    # convert it to an implicit and
    e = BooleanRuleExpression(OPERATORS.AND_ALL, None, exprs, None)
    return raw_evaluate_expression(e, pattern_ids_to_pattern_matches, flags)


def PatternMatchMock(
    start: int, end: int, metavars: Optional[Dict[str, Any]] = None
) -> MagicMock:
    mock = MagicMock()
    range_property = PropertyMock(return_value=Range(start, end, metavars or {}))
    type(mock).range = range_property
    metavars_property = PropertyMock(return_value=metavars)
    type(mock).metavars = metavars_property
    return mock


def RuleExpr(
    operator: Operator,
    fake_pattern_name: str,
    children: Optional[List[BooleanRuleExpression]] = None,
) -> BooleanRuleExpression:
    return BooleanRuleExpression(
        operator, PatternId(fake_pattern_name), children, "fake-pattern-text-here"
    )


def testA() -> None:
    """

    TODO: what about nested booleans?

    let pattern1 = subprocess.Popen($X, safeflag=True)
    let pattern2 =  subprocess.Popen($X, ...)

        import subprocess
        subprocess.Popen(subprocess.Popen(safeFlag=True))
        ------------------------------------------------- P2        R1
                         ------------------------------- P1, P2     R2

    and-not P1 --> remove all ranges equal to P1 exactly (removes R2)
    and P2 --> remove all ranges that don't equal P2 (R1 stays)
    OUTPUT: R1

    """
    results = {
        PatternId("pattern1"): [PatternMatchMock(30, 100)],
        PatternId("pattern2"): [PatternMatchMock(0, 100), PatternMatchMock(30, 100)],
    }
    expression = [
        RuleExpr(OPERATORS.AND_NOT, "pattern1"),
        RuleExpr(OPERATORS.AND, "pattern2"),
    ]
    result = evaluate_expression(expression, results)
    assert result == set([Range(0, 100, {})]), f"{result}"


def testB() -> None:
    """
    Our algebra needs to express "AND-INSIDE" as opposed to "AND-NOT-INSIDE", so that cases
    like this one can still fire (we explicitly don't want to ignore the nested expression
    just because it's inside).

        let pattern1 = subprocess.Popen($X, safeflag=True)
        let pattern2 = subprocess.Popen($X, ...)
        let pattern3 = importlib.Popen($X, ...)


        import subprocess
        subprocess.Popen(subprocess.Popen(bad), safeFlag=True)
        ----------------------------------------------------- P1, P2        R1
                         ---------------------                P2            R2

        and-not P1 --> remove all ranges == P1 exactly (R1 is removed)
        and-or (P2, P3) -->  remove any ranges not exactly == P2 or P3. (R2 stays)
        OUTPUT: R2
    """
    results = {
        PatternId("pattern1"): [PatternMatchMock(0, 100)],
        PatternId("pattern2"): [PatternMatchMock(30, 70), PatternMatchMock(0, 100)],
        PatternId("pattern3"): [],
    }
    expression = [
        RuleExpr(OPERATORS.AND_NOT, "pattern1"),
        BooleanRuleExpression(
            OPERATORS.AND_EITHER,
            None,
            [RuleExpr(OPERATORS.AND, "pattern2"), RuleExpr(OPERATORS.AND, "pattern3")],
        ),
    ]
    result = evaluate_expression(expression, results)
    assert result == set([Range(30, 70, {})]), f"{result}"


def testC() -> None:
    """
        let pattern1 = subprocess.Popen($X, safeflag=True)
        let pattern2 = subprocess.Popen($X, ...)
        let pattern4 = def __eq__(): \n...

        def __eq__():
            import subprocess
            subprocess.Popen(subprocess.Popen(bad), safeFlag=True)
            -----------------------------------------------------   P1, P2      R1
                            ---------------------                   P2          R2
        ----------------------------------------------------------  P4          R3

        and-inside P4 --> remove all ranges that are not enclosed by P4. Now only ranges inside or equal to P4 are left (all ranges remain)
        and-not P1 --> remove all ranges == P1. Now only ranges that don't have P1 remain (R2, R3 remain)
        and P2 --> remove all ranges not == P2. R2 remains.
        OUTPUT: R2
    """
    results = {
        PatternId("pattern1"): [PatternMatchMock(100, 1000)],
        PatternId("pattern2"): [
            PatternMatchMock(100, 1000),
            PatternMatchMock(200, 300),
        ],
        PatternId("pattern4"): [PatternMatchMock(0, 1000)],
    }
    expression = [
        RuleExpr(OPERATORS.AND_INSIDE, "pattern4"),
        RuleExpr(OPERATORS.AND_NOT, "pattern1"),
        RuleExpr(OPERATORS.AND, "pattern2"),
    ]
    result = evaluate_expression(expression, results)
    assert result == set([Range(200, 300, {})]), f"{result}"


def testD() -> None:
    """
        let pattern1 = subprocess.Popen($X, safeflag=True)
        let pattern2 = subprocess.Popen($X, ...)
        let pattern4 = def __eq__(): \n...

        def __eq__():
            import subprocess
            subprocess.Popen(subprocess.Popen(bad), safeFlag=True)
            -----------------------------------------------------   P1, P2      R1
                            ---------------------                   P2          R2
        ----------------------------------------------------------  P4          R3

        and-not-inside P4 --> remove all ranges that are not enclosed by P4. Now only ranges inside or equal to P4 are left (no ranges remain)
        and-not P1 --> no effect
        and P2 --> no effect
        OUTPUT: []
    """
    results = {
        PatternId("pattern1"): [PatternMatchMock(100, 1000)],
        PatternId("pattern2"): [
            PatternMatchMock(100, 1000),
            PatternMatchMock(200, 300),
        ],
        PatternId("pattern4"): [PatternMatchMock(0, 1000)],
    }
    expression = [
        RuleExpr(OPERATORS.AND_NOT_INSIDE, "pattern4"),
        RuleExpr(OPERATORS.AND_NOT, "pattern1"),
        RuleExpr(OPERATORS.AND, "pattern2"),
    ]
    result = evaluate_expression(expression, results)
    assert result == set([]), f"{result}"


def testE() -> None:
    """
    let pattern1 = bad(...)
    let pattern2 = def __eq__(): \n...
    let pattern3 = def normal(): \n...


        0-100 def __eq__():
        100-200    bad()

        200-300 def normal():
        300-400     bad(bad())
        400-500     def __eq__():
        500-600         bad()
    """

    """
        and-inside P3
        and-not-inside P2
        and P1
        OUTPUT: [300-400], [350-400]
    """
    results = {
        PatternId("pattern1"): [
            PatternMatchMock(100, 200),
            PatternMatchMock(300, 400),
            PatternMatchMock(350, 400),
            PatternMatchMock(500, 600),
        ],
        PatternId("pattern2"): [PatternMatchMock(0, 200), PatternMatchMock(400, 600)],
        PatternId("pattern3"): [PatternMatchMock(200, 600)],
    }
    expression = [
        RuleExpr(OPERATORS.AND_INSIDE, "pattern3"),
        RuleExpr(OPERATORS.AND_NOT_INSIDE, "pattern2"),
        RuleExpr(OPERATORS.AND, "pattern1"),
    ]
    result = evaluate_expression(expression, results)
    assert result == set([Range(300, 400, {}), Range(350, 400, {})]), f"{result}"

    """
        and-inside P2
        and-not-inside P3
        and P1
        OUTPUT: [100-200]
    """
    expression = [
        RuleExpr(OPERATORS.AND_INSIDE, "pattern2"),
        RuleExpr(OPERATORS.AND_NOT_INSIDE, "pattern3"),
        RuleExpr(OPERATORS.AND, "pattern1"),
    ]
    result = evaluate_expression(expression, results)
    assert result == set([Range(100, 200, {})]), f"{result}"

    """
        and-inside P1
        OUTPUT: [100-200, 300-400, 350-400, 500-600]
    """
    expression = [RuleExpr(OPERATORS.AND_INSIDE, "pattern1")]
    result = evaluate_expression(expression, results)
    assert result == set(
        [
            Range(100, 200, {}),
            Range(300, 400, {}),
            Range(350, 400, {}),
            Range(500, 600, {}),
        ]
    ), f"{result}"

    """
        and-inside-noteq P1
        and P1
        OUTPUT: [350-400]
    """

    # expression = [(OPERATORS.AND_INSIDE_NOT_EQ, ["pattern1"]),
    #              (OPERATORS.AND, ["pattern1"])]
    # result = evaluate_expression(expression, results)
    # assert result == set([Range(350, 400)]), f"{result}"
    # TODO


def testF() -> None:
    """Nested boolean expressions

    let pattern1 = bad(..., x=1)
    let pattern2 = bad(..., y=2)
    let pattern3 = def normal(): \n...
    let pattern4 = def unusual(): \n...

    We want to find (ONLY inside P3), either:
        P2 inside P4
      or
        P1 not-inside P4

    example:

    000-100    def normal():
    100-200        bad(x = 1) # P1 not-inside-P4
    200-300        bad(y = 2) # no match
    300-400        def unusual():
    400-500            bad(x = 1) # no match
    500-600            bad(y = 2) # P2 inside P4
    600-700        def regular():
    700-800            bad(x = 1) # P1 not-inside P4
    800-900            bad(y = 2) # no-match

        pattern-inside P3
        and-either:
           - patterns:
            - pattern-inside P4
            - and P2
          - patterns:
            - pattern-not-inside P4
            - and P1

        OUTPUT: [500-600], [700-800]
    """
    results = {
        PatternId("pattern1"): [
            PatternMatchMock(100, 200),
            PatternMatchMock(400, 500),
            PatternMatchMock(700, 800),
        ],
        PatternId("pattern2"): [
            PatternMatchMock(200, 300),
            PatternMatchMock(500, 600),
            PatternMatchMock(800, 900),
        ],
        PatternId("pattern3"): [PatternMatchMock(0, 900)],
        PatternId("pattern4"): [PatternMatchMock(300, 600)],
    }

    subexpression1 = [
        RuleExpr(OPERATORS.AND_INSIDE, "pattern4"),
        RuleExpr(OPERATORS.AND, "pattern2"),
    ]
    subexpression2 = [
        RuleExpr(OPERATORS.AND_NOT_INSIDE, "pattern4"),
        RuleExpr(OPERATORS.AND, "pattern1"),
    ]
    expression = [
        RuleExpr(OPERATORS.AND_INSIDE, "pattern3"),
        BooleanRuleExpression(
            OPERATORS.AND_EITHER,
            None,
            [
                BooleanRuleExpression(OPERATORS.AND_ALL, None, subexpression1),
                BooleanRuleExpression(OPERATORS.AND_ALL, None, subexpression2),
            ],
        ),
    ]
    result = evaluate_expression(expression, results)
    assert result == set(
        [Range(100, 200, {}), Range(500, 600, {}), Range(700, 800, {})]
    ), f"{result}"

    # TODO test and-all (`patterns` subkey)


def test_exprs() -> None:
    subexpression1 = [
        BooleanRuleExpression(OPERATORS.AND_INSIDE, PatternId("pattern4"), None, "p4"),
        BooleanRuleExpression(OPERATORS.AND, PatternId("pattern2"), None, "p2"),
    ]
    subexpression2 = [
        BooleanRuleExpression(
            OPERATORS.AND_NOT_INSIDE, PatternId("pattern4"), None, "p4"
        ),
        BooleanRuleExpression(OPERATORS.AND, PatternId("pattern1"), None, "p1"),
    ]
    expression = BooleanRuleExpression(
        OPERATORS.AND_ALL,
        None,
        [
            BooleanRuleExpression(
                OPERATORS.AND_INSIDE, PatternId("pattern3"), None, "p3"
            ),
            BooleanRuleExpression(
                OPERATORS.AND_EITHER,
                None,
                [
                    BooleanRuleExpression(
                        OPERATORS.AND_ALL, PatternId("someid"), subexpression1
                    ),
                    BooleanRuleExpression(
                        OPERATORS.AND_ALL, PatternId("someid2"), subexpression2
                    ),
                ],
            ),
        ],
    )
    flat = list(enumerate_patterns_in_boolean_expression(expression))
    # print(flat)

    expected = [
        BooleanRuleExpression(OPERATORS.AND_ALL, None, None, None),
        BooleanRuleExpression(OPERATORS.AND_INSIDE, PatternId("pattern3"), None, "p3"),
        BooleanRuleExpression(OPERATORS.AND_EITHER, None, None, None),
        BooleanRuleExpression(OPERATORS.AND_ALL, None, None, None),
        BooleanRuleExpression(OPERATORS.AND_INSIDE, PatternId("pattern4"), None, "p4"),
        BooleanRuleExpression(OPERATORS.AND, PatternId("pattern2"), None, "p2"),
        BooleanRuleExpression(OPERATORS.AND_ALL, None, None, None),
        BooleanRuleExpression(
            OPERATORS.AND_NOT_INSIDE, PatternId("pattern4"), None, "p4"
        ),
        BooleanRuleExpression(OPERATORS.AND, PatternId("pattern1"), None, "p1"),
    ]

    assert flat == expected, f"flat: {flat}"


def test_build_exprs() -> None:
    base_rule: Dict[str, Any] = {
        "id": "test-id",
        "message": "test message",
        "languages": ["python"],
        "severity": "ERROR",
    }
    rules: List[Dict[str, Any]] = [
        {**base_rule, **{"pattern": "test(...)"}},
        {**base_rule, **{"patterns": [{"pattern": "test(...)"}]}},
        {**base_rule, **{"pattern-either": [{"pattern": "test(...)"}]}},
    ]

    results = [Rule.from_json(rule).expression for rule in rules]
    base_expected = [
        BooleanRuleExpression(OPERATORS.AND, PatternId(".0"), None, "test(...)")
    ]
    expected = [
        BooleanRuleExpression(OPERATORS.AND, PatternId("test-id"), None, "test(...)"),
        BooleanRuleExpression(OPERATORS.AND_ALL, None, base_expected, None),
        BooleanRuleExpression(OPERATORS.AND_EITHER, None, base_expected, None),
    ]

    assert results == expected


def test_evaluate_python() -> None:
    """Test evaluating the subpattern `where-python: <python_expression>`,
    in which a rule can provide an arbitrary Python expression that will be
    evaluated against the currently matched metavariables.

    NOTE: the Python expression must evaluate to True or False.

    NOTE: Assume patterns are applied in the order specified, top to bottom.

    This is implementing: https://github.com/returntocorp/semgrep/issues/101.

        let allExecs = exec($X)
        let filteredExecs = where-python: "vars['$X'].startswith('cmd')"

        000-100   var exec = require('child_process').exec;

        100-200   var cmd_pattern = "user_input";
        200-300   var other_pattern = "hardcoded_string";

        300-400   // should match
        400-500   exec(cmd_pattern, function(error, stdout, stderr){
        500-600       console.log(stdout);
        600-700   });

        700-800   // should not match
        800-900   exec(other_pattern, function(error, stdout, stderr){
        900-1000      console.log(stdout);
        1100-1200 });

        patterns:
            pattern: exec($X)
            where-python: "vars['$X'].startswith('cmd')"


        OUTPUT: [400-500]
    """
    results = {
        PatternId("all_execs"): [
            PatternMatchMock(400, 500, {"$X": {"abstract_content": "cmd_pattern"}}),
            PatternMatchMock(800, 900, {"$X": {"abstract_content": "other_pattern"}}),
        ]
    }

    expression = [
        BooleanRuleExpression(OPERATORS.AND, PatternId("all_execs"), None, "all_execs"),
        BooleanRuleExpression(
            OPERATORS.WHERE_PYTHON,
            PatternId("p1"),
            None,
            "vars['$X'].startswith('cmd')",
        ),
    ]

    result = evaluate_expression(expression, results, flags={RCE_RULE_FLAG: True})
    assert result == set([Range(400, 500, {})]), f"{result}"



#!/usr/bin/env python3
import sys

from semgrep.cli import cli
from semgrep.error import SemgrepException
from semgrep.util import print_error


def main() -> None:
    try:
        cli()
    # Catch custom exceptions, output the right message and exit.
    # Note: this doesn't catch all Exceptions and lets them bubble up.
    except SemgrepException as e:
        if e.msg:
            print_error(e.msg)
        sys.exit(e.code)


if __name__ == "__main__":
    main()



import subprocess
from pathlib import Path

import pytest


pytestmark = pytest.mark.qa


def test_semgrep_rules_repo(run_semgrep_in_tmp):
    subprocess.check_output(
        ["git", "clone", "--depth=1", "https://github.com/returntocorp/semgrep-rules"]
    )
    subprocess.check_output(["python3", "-m", "semgrep", "--generate-config"])
    subprocess.check_output(
        [
            "python3",
            "-m",
            "semgrep",
            "--dangerously-allow-arbitrary-code-execution-from-rules",
            "--strict",
            "--test",
            "--test-ignore-todo",
            "semgrep-rules",
        ]
    )
    subprocess.check_output(["python3", "-m", "semgrep", "--validate", "semgrep-rules"])



from subprocess import CalledProcessError

import pytest


@pytest.mark.parametrize("filename", ["good", "good_info_severity", "good_metadata"])
def test_rule_parser__success(run_semgrep_in_tmp, snapshot, filename):
    run_semgrep_in_tmp(f"rules/syntax/{filename}.yaml")


@pytest.mark.parametrize(
    "filename",
    ["bad1", "bad2", "bad3", "bad4", "badpattern", "badpaths1", "badpaths2"],
)
def test_rule_parser__failure(run_semgrep_in_tmp, snapshot, filename):
    with pytest.raises(CalledProcessError) as excinfo:
        run_semgrep_in_tmp(f"rules/syntax/{filename}.yaml")
    assert excinfo.value.returncode != 0


@pytest.mark.parametrize(
    "filename",
    [
        pytest.param(
            "bad1",
            marks=pytest.mark.xfail(
                reason="https://github.com/returntocorp/semgrep/issues/678"
            ),
        ),
        pytest.param(
            "bad2", marks=pytest.mark.xfail(reason="prints set with arbitrary order")
        ),
        "bad3",
        "bad4",
        "badpattern",
        "badpaths1",
        "badpaths2",
    ],
)
def test_rule_parser__failure__error_messages(run_semgrep_in_tmp, snapshot, filename):
    with pytest.raises(CalledProcessError) as excinfo:
        run_semgrep_in_tmp(f"rules/syntax/{filename}.yaml", stderr=True)
    snapshot.assert_match(excinfo.value.output, "error.txt")



import sys
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import semgrep.config_resolver
from semgrep.autofix import apply_fixes
from semgrep.constants import DEFAULT_CONFIG_FILE
from semgrep.constants import ID_KEY
from semgrep.constants import OutputFormat
from semgrep.constants import RULES_KEY
from semgrep.core_runner import CoreRunner
from semgrep.output import build_output
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch
from semgrep.semgrep_types import InvalidRuleSchema
from semgrep.semgrep_types import YAML_ALL_VALID_RULE_KEYS
from semgrep.semgrep_types import YAML_MUST_HAVE_KEYS
from semgrep.util import debug_print
from semgrep.util import FINDINGS_EXIT_CODE
from semgrep.util import INVALID_CODE_EXIT_CODE
from semgrep.util import is_url
from semgrep.util import MISSING_CONFIG_EXIT_CODE
from semgrep.util import print_error
from semgrep.util import print_error_exit
from semgrep.util import print_msg

MISSING_RULE_ID = "no-rule-id"


def validate_single_rule(config_id: str, rule: Dict[str, Any]) -> bool:
    """
        Validate that a rule dictionary contains all necessary keys
        and can be correctly parsed
    """
    rule_id_err_msg = f'{rule.get("id", MISSING_RULE_ID)}'
    rule_keys = set(rule.keys())
    if not rule_keys.issuperset(YAML_MUST_HAVE_KEYS):
        missing_keys = YAML_MUST_HAVE_KEYS - rule_keys
        print_error(
            f"{config_id} is missing required keys {missing_keys} at rule id {rule_id_err_msg}"
        )
        return False
    if not rule_keys.issubset(YAML_ALL_VALID_RULE_KEYS):
        extra_keys = rule_keys - YAML_ALL_VALID_RULE_KEYS
        print_error(
            f"{config_id} has invalid rule key {extra_keys} at rule id {rule_id_err_msg}, can only have: {YAML_ALL_VALID_RULE_KEYS}"
        )
        return False
    try:
        _ = Rule.from_json(rule).expression
    except InvalidRuleSchema as ex:
        print_error(
            f"{config_id}: inside rule id {rule_id_err_msg}, pattern fields can't look like this: {ex}"
        )
        return False
    try:
        _ = Rule.from_json(rule).globs
    except InvalidRuleSchema as ex:
        print_error(
            f"{config_id}: inside rule id {rule_id_err_msg}, path fields can't look like this: {ex}"
        )
        return False

    return True


def validate_configs(
    configs: Dict[str, Optional[Dict[str, Any]]]
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
        Take configs and separate into valid and invalid ones
    """
    errors = {}
    valid = {}
    for config_id, config in configs.items():
        if not config:
            errors[config_id] = config
            continue
        if RULES_KEY not in config:
            print_error(f"{config_id} is missing `{RULES_KEY}` as top-level key")
            errors[config_id] = config
            continue
        rules = config.get(RULES_KEY)
        valid_rules = []
        invalid_rules = []
        for rule in rules:  # type: ignore
            if validate_single_rule(config_id, rule):
                valid_rules.append(rule)
            else:
                invalid_rules.append(rule)

        if invalid_rules:
            errors[config_id] = {**config, "rules": invalid_rules}
        if valid_rules:
            valid[config_id] = {**config, "rules": valid_rules}
    return valid, errors


def safe_relative_to(a: Path, b: Path) -> Path:
    try:
        return a.relative_to(b)
    except ValueError:
        # paths had no common prefix; not possible to relativize
        return a


def convert_config_id_to_prefix(config_id: str) -> str:
    at_path = Path(config_id)
    at_path = safe_relative_to(at_path, Path.cwd())

    prefix = ".".join(at_path.parts[:-1]).lstrip("./").lstrip(".")
    if len(prefix):
        prefix += "."
    return prefix


def rename_rule_ids(valid_configs: Dict[str, Any]) -> Dict[str, Any]:
    transformed = {}
    for config_id, config in valid_configs.items():
        rules = config.get(RULES_KEY, [])
        transformed_rules = [
            {
                **rule,
                ID_KEY: f"{convert_config_id_to_prefix(config_id)}{rule.get(ID_KEY, MISSING_RULE_ID)}",
            }
            for rule in rules
        ]
        transformed[config_id] = {**config, RULES_KEY: transformed_rules}
    return transformed


### Handle output


def post_output(output_url: str, output: str) -> None:
    import requests  # here for faster startup times

    print_msg(f"posting to {output_url}...")
    try:
        r = requests.post(output_url, data=output, timeout=10)
        debug_print(f"posted to {output_url} and got status_code:{r.status_code}")
    except requests.exceptions.Timeout:
        print_error_exit(f"posting output to {output_url} timed out")


def save_output(destination: str, output: str) -> None:
    if is_url(destination):
        post_output(destination, output)
    else:
        if Path(destination).is_absolute():
            save_path = Path(destination)
        else:
            base_path = semgrep.config_resolver.get_base_path()
            save_path = base_path.joinpath(destination)
        # create the folders if not exists
        save_path.parent.mkdir(parents=True, exist_ok=True)
        with save_path.open(mode="w") as fout:
            fout.write(output)


def get_config(generate_config: bool, pattern: str, lang: str, config: str) -> Any:
    # first check if user asked to generate a config
    if generate_config:
        semgrep.config_resolver.generate_config()

    # let's check for a pattern
    elif pattern:
        # and a language
        if not lang:
            print_error_exit("language must be specified when a pattern is passed")

        # TODO for now we generate a manual config. Might want to just call semgrep -e ... -l ...
        configs = semgrep.config_resolver.manual_config(pattern, lang)
    else:
        # else let's get a config. A config is a dict from config_id -> config. Config Id is not well defined at this point.
        configs = semgrep.config_resolver.resolve_config(config)

    # if we can't find a config, use default r2c rules
    if not configs:
        print_error_exit(
            f"No config given and {DEFAULT_CONFIG_FILE} was not found. Try running with --help to debug or if you want to download a default config, try running with --config r2c"
        )

    # let's split our configs into valid and invalid configs.
    # It's possible that a config_id exists in both because we check valid rules and invalid rules
    # instead of just hard failing for that config if mal-formed
    valid_configs, invalid_configs = validate_configs(configs)
    return valid_configs, invalid_configs


def flatten_configs(transformed_configs: Dict[str, Any]) -> List[Rule]:
    return [
        Rule.from_json(rule)
        for config in transformed_configs.values()
        for rule in config.get(RULES_KEY, [])
    ]


def should_exclude_this_path(path: Path) -> bool:
    """
        Return true if "test" or "example" are anywhere in path
    """
    return any("test" in p or "example" in p for p in path.parts)


def main(
    target: List[str],
    pattern: str,
    lang: str,
    config: str,
    generate_config: bool,
    no_rewrite_rule_ids: bool,
    jobs: int,
    include: List[str],
    include_dir: List[str],
    exclude: List[str],
    exclude_dir: List[str],
    exclude_tests: bool,
    json_format: bool,
    sarif: bool,
    output_destination: str,
    quiet: bool,
    strict: bool,
    exit_on_error: bool,
    autofix: bool,
    dangerously_allow_arbitrary_code_execution_from_rules: bool,
) -> str:
    # get the proper paths for targets i.e. handle base path of /home/repo when it exists in docker
    targets = semgrep.config_resolver.resolve_targets(target)
    valid_configs, invalid_configs = get_config(generate_config, pattern, lang, config)

    output_format = OutputFormat.TEXT
    if json_format:
        output_format = OutputFormat.JSON
    elif sarif:
        output_format = OutputFormat.SARIF

    if invalid_configs and strict:
        print_error_exit(
            f"run with --strict and there were {len(invalid_configs)} errors loading configs",
            MISSING_CONFIG_EXIT_CODE,
        )

    if not no_rewrite_rule_ids:
        # re-write the configs to have the hierarchical rule ids
        valid_configs = rename_rule_ids(valid_configs)

    # extract just the rules from valid configs
    all_rules = flatten_configs(valid_configs)

    if not pattern:
        plural = "s" if len(valid_configs) > 1 else ""
        config_id_if_single = (
            list(valid_configs.keys())[0] if len(valid_configs) == 1 else ""
        )
        invalid_msg = (
            f"({len(invalid_configs)} config files were invalid)"
            if len(invalid_configs)
            else ""
        )
        debug_print(
            f"running {len(all_rules)} rules from {len(valid_configs)} config{plural} {config_id_if_single} {invalid_msg}"
        )

        if len(valid_configs) == 0:
            print_error_exit(
                f"no valid configuration file found ({len(invalid_configs)} configs were invalid)",
                MISSING_CONFIG_EXIT_CODE,
            )

    # actually invoke semgrep
    rule_matches_by_rule, semgrep_errors = CoreRunner(
        allow_exec=dangerously_allow_arbitrary_code_execution_from_rules,
        jobs=jobs,
        exclude=exclude,
        include=include,
        exclude_dir=exclude_dir,
        include_dir=include_dir,
    ).invoke_semgrep(targets, all_rules)

    if exclude_tests:
        ignored_in_tests = 0
        filtered_findings_by_rule = {}
        for rule, rule_matches in rule_matches_by_rule.items():
            filtered = []
            for rule_match in rule_matches:
                if should_exclude_this_path(Path(rule_match.path)):
                    ignored_in_tests += 1
                else:
                    filtered.append(rule_match)
            filtered_findings_by_rule[rule] = filtered
        rule_matches_by_rule = filtered_findings_by_rule
        if ignored_in_tests > 0:
            print_error(
                f"warning: ignored {ignored_in_tests} results in tests due to --exclude-tests option"
            )

    for finding in semgrep_errors:
        print_error(f"semgrep: {finding['path']}: {finding['check_id']}")

    if strict and len(semgrep_errors):
        print_error_exit(
            f"run with --strict and {len(semgrep_errors)} errors occurred during semgrep run; exiting",
            INVALID_CODE_EXIT_CODE,
        )

    output = handle_output(
        rule_matches_by_rule,
        semgrep_errors,
        output_format,
        quiet,
        output_destination,
        exit_on_error,
    )
    if autofix:
        apply_fixes(rule_matches_by_rule)

    return output


def handle_output(
    rule_matches_by_rule: Dict[Rule, List[RuleMatch]],
    semgrep_errors: List[Any],
    output_format: OutputFormat,
    quiet: bool = False,
    output_destination: Optional[str] = None,
    exit_on_error: bool = False,
) -> str:
    rules = frozenset(rule_matches_by_rule.keys())
    rule_matches = [
        match
        for matches_of_one_rule in rule_matches_by_rule.values()
        for match in matches_of_one_rule
    ]

    output = build_output(
        rule_matches, rules, semgrep_errors, output_format, not output_destination
    )
    if not quiet:
        if output:
            print(output)

    if output_destination:
        save_output(output_destination, output)

    if exit_on_error and any(match.should_fail_run for match in rule_matches):
        sys.exit(FINDINGS_EXIT_CODE)

    # TODO tests still rely on this but it is unnecessary
    return output



from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Set

from semgrep.constants import RCE_RULE_FLAG
from semgrep.pattern_match import PatternMatch
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch
from semgrep.semgrep_types import BooleanRuleExpression
from semgrep.semgrep_types import OPERATORS
from semgrep.semgrep_types import pattern_names_for_operator
from semgrep.semgrep_types import PatternId
from semgrep.semgrep_types import Range
from semgrep.util import debug_print
from semgrep.util import flatten
from semgrep.util import NEED_ARBITRARY_CODE_EXEC_EXIT_CODE
from semgrep.util import print_error
from semgrep.util import print_error_exit


def _evaluate_single_expression(
    expression: BooleanRuleExpression,
    pattern_ids_to_pattern_matches: Dict[PatternId, List[PatternMatch]],
    ranges_left: Set[Range],
    flags: Optional[Dict[str, Any]] = None,
) -> Set[Range]:

    assert expression.pattern_id, f"<internal error: expected pattern id: {expression}>"
    results_for_pattern = [
        x.range for x in pattern_ids_to_pattern_matches.get(expression.pattern_id, [])
    ]

    if expression.operator == OPERATORS.AND:
        # remove all ranges that don't equal the ranges for this pattern
        return ranges_left.intersection(results_for_pattern)
    elif expression.operator == OPERATORS.AND_NOT:
        # remove all ranges that DO equal the ranges for this pattern
        # difference_update = Remove all elements of another set from this set.
        return ranges_left.difference(results_for_pattern)
    elif expression.operator == OPERATORS.AND_INSIDE:
        # remove all ranges (not enclosed by) or (not equal to) the inside ranges
        output_ranges = set()
        for arange in ranges_left:
            for keep_inside_this_range in results_for_pattern:
                is_enclosed = keep_inside_this_range.is_enclosing_or_eq(arange)
                # print(
                #    f'candidate range is {arange}, needs to be `{operator}` {keep_inside_this_range}; keep?: {keep}')
                if is_enclosed:
                    output_ranges.add(arange)
                    break  # found a match, no need to keep going
        debug_print(f"after filter `{expression.operator}`: {output_ranges}")
        return output_ranges
    elif expression.operator == OPERATORS.AND_NOT_INSIDE:
        # remove all ranges enclosed by or equal to
        output_ranges = ranges_left.copy()
        for arange in ranges_left:
            for keep_inside_this_range in results_for_pattern:
                if keep_inside_this_range.is_enclosing_or_eq(arange):
                    output_ranges.remove(arange)
                    break
        debug_print(f"after filter `{expression.operator}`: {output_ranges}")
        return output_ranges
    elif expression.operator == OPERATORS.WHERE_PYTHON:
        if not flags or flags[RCE_RULE_FLAG] != True:
            print_error_exit(
                f"at least one rule needs to execute arbitrary code; this is dangerous! if you want to continue, enable the flag: {RCE_RULE_FLAG}",
                NEED_ARBITRARY_CODE_EXEC_EXIT_CODE,
            )
        assert expression.operand, "must have operand for this operator type"

        output_ranges = set()
        # Look through every range that hasn't been filtered yet
        for pattern_match in list(flatten(pattern_ids_to_pattern_matches.values())):
            # Only need to check where-python clause if the range hasn't already been filtered

            if pattern_match.range in ranges_left:
                debug_print(
                    f"WHERE is {expression.operand}, metavars: {pattern_match.metavars}"
                )
                if _where_python_statement_matches(
                    expression.operand, pattern_match.metavars
                ):
                    output_ranges.add(pattern_match.range)
        debug_print(f"after filter `{expression.operator}`: {output_ranges}")
        return output_ranges
    elif expression.operator == OPERATORS.REGEX:
        # remove all ranges that don't equal the ranges for this pattern
        return ranges_left.intersection(results_for_pattern)
    else:
        raise NotImplementedError(f"unknown operator {expression.operator}")


# Given a `where-python` expression as a string and currently matched metavars,
# return whether the expression matches as a boolean
def _where_python_statement_matches(
    where_expression: str, metavars: Dict[str, Any]
) -> bool:
    # TODO: filter out obvious dangerous things here
    global output
    output = None  # type: ignore

    # HACK: we're executing arbitrary Python in the where-python,
    # be careful my friend
    vars = {k: v["abstract_content"] for k, v in metavars.items()}
    try:
        exec(f"global output; output = {where_expression}")
    except Exception as ex:
        print_error(
            f"error evaluating a where-python expression: `{where_expression}`: {ex}"
        )

    if type(output) != type(True):  # type: ignore
        print_error_exit(
            f"python where expression needs boolean output but got: {output} for {where_expression}"  # type: ignore
        )
    return output == True  # type: ignore


def group_by_pattern_id(
    pattern_matches: List[PatternMatch],
) -> Dict[PatternId, List[PatternMatch]]:
    by_id: Dict[PatternId, List[PatternMatch]] = {}
    for pattern_match in pattern_matches:
        by_id.setdefault(pattern_match.id, []).append(pattern_match)
    return by_id


def safe_relative_to(a: Path, b: Path) -> Path:
    try:
        return a.relative_to(b)
    except ValueError:
        # paths had no common prefix; not possible to relativize
        return a


def should_exclude_this_path(path: Path) -> bool:
    return any("test" in p or "example" in p for p in path.parts)


def evaluate(
    rule: Rule, pattern_matches: List[PatternMatch], allow_exec: bool
) -> List[RuleMatch]:
    """
        Takes a Rule and list of pattern matches from a single file and
        handles the boolean expression evaluation of the Rule's patterns
        Returns a list of RuleMatches.
    """
    output = []
    pattern_ids_to_pattern_matches = group_by_pattern_id(pattern_matches)
    debug_print(str(pattern_ids_to_pattern_matches))
    valid_ranges_to_output = evaluate_expression(
        rule.expression,
        pattern_ids_to_pattern_matches,
        flags={RCE_RULE_FLAG: allow_exec},
    )

    # only output matches which are inside these offsets!
    debug_print(f"compiled result {valid_ranges_to_output}")
    debug_print("-" * 80)
    for pattern_match in pattern_matches:
        if pattern_match.range in valid_ranges_to_output:
            message = interpolate_message_metavariables(rule, pattern_match)
            fix = interpolate_fix_metavariables(rule, pattern_match)
            rule_match = RuleMatch(
                rule.id,
                pattern_match,
                message=message,
                metadata=rule.metadata,
                severity=rule.severity,
                fix=fix,
            )
            output.append(rule_match)

    return output


def interpolate_message_metavariables(rule: Rule, pattern_match: PatternMatch) -> str:
    msg_text = rule.message
    for metavar, contents in pattern_match.metavars.items():
        msg_text = msg_text.replace(metavar, contents.get("abstract_content", ""))
    return msg_text


def interpolate_fix_metavariables(
    rule: Rule, pattern_match: PatternMatch
) -> Optional[str]:
    fix_str = rule.fix
    if fix_str is None:
        return None
    for metavar, contents in pattern_match.metavars.items():
        fix_str = fix_str.replace(metavar, contents.get("abstract_content", ""))
    return fix_str


def evaluate_expression(
    expression: BooleanRuleExpression,
    pattern_ids_to_pattern_matches: Dict[PatternId, List[PatternMatch]],
    flags: Optional[Dict[str, Any]] = None,
) -> Set[Range]:
    ranges_left = set(
        [x.range for x in flatten(pattern_ids_to_pattern_matches.values())]
    )
    return _evaluate_expression(
        expression, pattern_ids_to_pattern_matches, ranges_left, flags
    )


def _evaluate_expression(
    expression: BooleanRuleExpression,
    pattern_ids_to_pattern_matches: Dict[PatternId, List[PatternMatch]],
    ranges_left: Set[Range],
    flags: Optional[Dict[str, Any]] = None,
) -> Set[Range]:
    if (
        expression.operator == OPERATORS.AND_EITHER
        or expression.operator == OPERATORS.AND_ALL
    ):
        assert (
            expression.children is not None
        ), f"{pattern_names_for_operator(OPERATORS.AND_EITHER)} or {pattern_names_for_operator(OPERATORS.AND_ALL)} must have a list of subpatterns"

        # recurse on the nested expressions
        if expression.operator == OPERATORS.AND_EITHER:
            # remove anything that does not equal one of these ranges
            evaluated_ranges = [
                _evaluate_expression(
                    expr, pattern_ids_to_pattern_matches, ranges_left.copy(), flags
                )
                for expr in expression.children
            ]
            ranges_left.intersection_update(flatten(evaluated_ranges))
        else:
            # chain intersection eagerly; intersect for every AND'ed child
            for expr in expression.children:
                remainining_ranges = _evaluate_expression(
                    expr, pattern_ids_to_pattern_matches, ranges_left.copy(), flags
                )
                ranges_left.intersection_update(remainining_ranges)

        debug_print(f"after filter `{expression.operator}`: {ranges_left}")
    else:
        assert (
            expression.children is None
        ), f"only `{pattern_names_for_operator(OPERATORS.AND_EITHER)}` or `{pattern_names_for_operator(OPERATORS.AND_ALL)}` expressions can have multiple subpatterns"
        ranges_left = _evaluate_single_expression(
            expression, pattern_ids_to_pattern_matches, ranges_left, flags
        )
    return ranges_left


def enumerate_patterns_in_boolean_expression(
    expr: BooleanRuleExpression,
) -> Iterable[BooleanRuleExpression]:
    """
    flatten a potentially nested expression
    """
    if expr.children is not None:
        # we need to preserve this parent of multiple children, but it has no corresponding pattern
        yield BooleanRuleExpression(expr.operator, None, None, None)
        # now yield all the children
        for c in expr.children:
            yield from enumerate_patterns_in_boolean_expression(c)
    else:
        yield expr



import json
from typing import Any
from typing import Dict
from typing import FrozenSet
from typing import Iterator
from typing import List

import colorama

from semgrep.constants import __VERSION__
from semgrep.constants import OutputFormat
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch


def color_line(
    line: str,
    line_number: int,
    start_line: int,
    start_col: int,
    end_line: int,
    end_col: int,
) -> str:
    start_color = 0 if line_number > start_line else start_col
    # column offset
    start_color = max(start_color - 1, 0)
    end_color = end_col if line_number >= end_line else len(line) + 1 + 1
    end_color = max(end_color - 1, 0)
    line = (
        line[:start_color]
        + colorama.Style.BRIGHT
        + line[start_color : end_color + 1]  # want the color to include the end_col
        + colorama.Style.RESET_ALL
        + line[end_color + 1 :]
    )
    return line


def finding_to_line(rule_match: RuleMatch, color_output: bool) -> Iterator[str]:
    path = rule_match.path
    start_line = rule_match.start.get("line")
    end_line = rule_match.end.get("line")
    start_col = rule_match.start.get("col")
    end_col = rule_match.end.get("col")
    if path:
        for i, line in enumerate(rule_match.lines):
            line = line.rstrip()
            line_number = ""
            if start_line:
                if color_output:
                    line = color_line(
                        line, start_line + i, start_line, start_col, end_line, end_col  # type: ignore
                    )
                    line_number = f"{colorama.Fore.GREEN}{start_line + i}{colorama.Style.RESET_ALL}"
                else:
                    line_number = f"{start_line + i}"

            yield f"{line_number}:{line}" if line_number else f"{line}"


def build_normal_output(
    rule_matches: List[RuleMatch], semgrep_errors: List[Any], color_output: bool
) -> Iterator[str]:
    RESET_COLOR = colorama.Style.RESET_ALL if color_output else ""
    GREEN_COLOR = colorama.Fore.GREEN if color_output else ""
    YELLOW_COLOR = colorama.Fore.YELLOW if color_output else ""
    RED_COLOR = colorama.Fore.RED if color_output else ""
    BLUE_COLOR = colorama.Fore.BLUE if color_output else ""

    last_file = None
    last_message = None
    for rule_match in sorted(rule_matches, key=lambda r: (r.path, r.id),):

        current_file = rule_match.path
        check_id = rule_match.id
        extra = rule_match.extra
        message = rule_match.message
        severity = rule_match.severity
        fix = rule_match.fix
        if last_file is None or last_file != current_file:
            if last_file is not None:
                yield ""
            yield f"{GREEN_COLOR}{current_file}{RESET_COLOR}"
            last_message = None
        # don't display the rule line if the check is empty
        if (
            check_id
            and check_id != "-"
            and (last_message is None or last_message != message)
        ):
            severity_prepend = ""
            if severity:
                if severity == "ERROR":
                    severity_prepend = f"{RED_COLOR}{severity} "
                elif severity == "WARNING":
                    severity_prepend = f"{YELLOW_COLOR}{severity} "
                else:
                    severity_prepend = f"{severity} "
            yield f"{severity_prepend}{YELLOW_COLOR}rule:{check_id}: {message}{RESET_COLOR}"

        last_file = current_file
        last_message = message
        yield from finding_to_line(rule_match, color_output)
        if fix:
            yield f"{BLUE_COLOR}autofix:{RESET_COLOR} {fix}"


def build_output_json(rule_matches: List[RuleMatch], semgrep_errors: List[Any]) -> str:
    # wrap errors under "data" entry to be compatible with
    # https://docs.r2c.dev/en/latest/api/output.html#errors
    output_json = {}
    output_json["results"] = [rm.to_json() for rm in rule_matches]
    output_json["errors"] = [
        {"data": e, "message": "SemgrepCoreRuntimeErrors"} for e in semgrep_errors
    ]
    return json.dumps(output_json)


def _sarif_tool_info() -> Dict[str, Any]:
    return {"name": "semgrep", "semanticVersion": __VERSION__}


def build_sarif_output(
    rule_matches: List[RuleMatch], rules: FrozenSet[Rule], semgrep_errors: List[Any]
) -> str:
    """
    Format matches in SARIF v2.1.0 formatted JSON.

    - written based on https://help.github.com/en/github/finding-security-vulnerabilities-and-errors-in-your-code/about-sarif-support-for-code-scanning
    - which links to this schema https://github.com/oasis-tcs/sarif-spec/blob/master/Schemata/sarif-schema-2.1.0.json
    - full spec is at https://docs.oasis-open.org/sarif/sarif/v2.1.0/cs01/sarif-v2.1.0-cs01.html
    """
    output_dict = {
        "$schema": "https://raw.githubusercontent.com/oasis-tcs/sarif-spec/master/Schemata/sarif-schema-2.1.0.json",
        "version": "2.1.0",
        "tool": {
            "driver": {
                **_sarif_tool_info(),
                "rules": [rule.to_sarif() for rule in rules],
            }
        },
        "results": [match.to_sarif() for match in rule_matches],
    }
    return json.dumps(output_dict)


def build_output(
    rule_matches: List[RuleMatch],
    rules: FrozenSet[Rule],
    semgrep_errors: List[Any],
    output_format: OutputFormat,
    color_output: bool,
) -> str:
    if output_format == OutputFormat.JSON:
        return build_output_json(rule_matches, semgrep_errors)
    elif output_format == OutputFormat.SARIF:
        return build_sarif_output(rule_matches, rules, semgrep_errors)
    elif output_format == OutputFormat.TEXT:
        return "\n".join(
            build_normal_output(rule_matches, semgrep_errors, color_output)
        )
    else:
        # https://github.com/python/mypy/issues/6366
        raise RuntimeError(f"Unhandled output format: {type(output_format).__name__}")
